@@ -26,6 +26,7 @@ typedef struct {
     PyObject *wrapper;
     PyObject *adapter;
     PyObject *bound_type;
+    PyObject *parent;
 } WraptFunctionWrapperObject;
 
 PyTypeObject WraptFunctionWrapperBase_Type;
@@ -1117,6 +1118,11 @@ static int WraptObjectProxy_set_wrapped(WraptObjectProxyObject *self,
       return -1;
     }
 
+    if (!value) {
+        PyErr_SetString(PyExc_TypeError, "__wrapped__ must be an object");
+        return -1;
+    }
+
     Py_INCREF(value);
     Py_DECREF(self->wrapped);
 
@@ -1178,21 +1184,21 @@ static PyObject *WraptObjectProxy_getattro(
 static int WraptObjectProxy_setattro(
         WraptObjectProxyObject *self, PyObject *name, PyObject *value)
 {
-    PyObject *self_prefix = NULL;
-    PyObject *attr_wrapped = NULL;
+    static PyObject *self_prefix = NULL;
+    static PyObject *attr_wrapped = NULL;
 
     PyObject *match = NULL;
 
+    if (!self_prefix) {
 #if PY_MAJOR_VERSION >= 3
-    self_prefix = PyUnicode_FromString("_self_");
+        self_prefix = PyUnicode_InternFromString("_self_");
 #else
-    self_prefix = PyString_FromString("_self_");
+        self_prefix = PyString_InternFromString("_self_");
 #endif
+    }
 
     match = PyEval_CallMethod(name, "startswith", "(O)", self_prefix);
 
-    Py_DECREF(self_prefix);
-
     if (match == Py_True) {
         Py_DECREF(match);
 
@@ -1201,11 +1207,13 @@ static int WraptObjectProxy_setattro(
 
     Py_XDECREF(match);
 
+    if (!attr_wrapped) {
 #if PY_MAJOR_VERSION >= 3
-    attr_wrapped = PyUnicode_FromString("__wrapped__");
+        attr_wrapped = PyUnicode_InternFromString("__wrapped__");
 #else
-    attr_wrapped = PyString_FromString("__wrapped__");
+        attr_wrapped = PyString_InternFromString("__wrapped__");
 #endif
+    }
 
     if (PyObject_RichCompareBool(name, attr_wrapped, Py_EQ) == 1) {
         Py_DECREF(attr_wrapped);
@@ -1213,8 +1221,6 @@ static int WraptObjectProxy_setattro(
         return PyObject_GenericSetAttr((PyObject *)self, name, value);
     }
 
-    Py_DECREF(attr_wrapped);
-
     if (!self->wrapped) {
       PyErr_SetString(PyExc_ValueError, "wrapper has not been initialised");
       return -1;
@@ -1416,6 +1422,7 @@ static PyObject *WraptFunctionWrapperBase_new(PyTypeObject *type,
     self->wrapper = NULL;
     self->adapter = NULL;
     self->bound_type = NULL;
+    self->parent = NULL;
 
     return (PyObject *)self;
 }
@@ -1424,7 +1431,7 @@ static PyObject *WraptFunctionWrapperBase_new(PyTypeObject *type,
 
 static int WraptFunctionWrapperBase_raw_init(WraptFunctionWrapperObject *self,
         PyObject *wrapped, PyObject *instance, PyObject *wrapper,
-         PyObject *adapter, PyObject *bound_type)
+         PyObject *adapter, PyObject *bound_type, PyObject *parent)
 {
     int result = 0;
 
@@ -1447,6 +1454,10 @@ static int WraptFunctionWrapperBase_raw_init(WraptFunctionWrapperObject *self,
         Py_INCREF(bound_type);
         Py_XDECREF(self->bound_type);
         self->bound_type = bound_type;
+
+        Py_INCREF(parent);
+        Py_XDECREF(self->parent);
+        self->parent = parent;
     }
 
     return result;
@@ -1462,17 +1473,19 @@ static int WraptFunctionWrapperBase_init(WraptFunctionWrapperObject *self,
     PyObject *wrapper = NULL;
     PyObject *adapter = Py_None;
     PyObject *bound_type = Py_None;
+    PyObject *parent = Py_None;
 
     static char *kwlist[] = { "wrapped", "instance", "wrapper",
-            "adapter", "bound_type", NULL };
+            "adapter", "bound_type", "parent", NULL };
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "OOO|OO:FunctionWrapperBase",
-            kwlist, &wrapped, &instance, &wrapper, &adapter, &bound_type)) {
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "OOO|OOO:FunctionWrapperBase",
+            kwlist, &wrapped, &instance, &wrapper, &adapter, &bound_type,
+            &parent)) {
         return -1;
     }
 
     return WraptFunctionWrapperBase_raw_init(self, wrapped, instance, wrapper,
-            adapter, bound_type);
+            adapter, bound_type, parent);
 }
 
 /* ------------------------------------------------------------------------- */
@@ -1486,6 +1499,7 @@ static int WraptFunctionWrapperBase_traverse(WraptFunctionWrapperObject *self,
     Py_VISIT(self->wrapper);
     Py_VISIT(self->adapter);
     Py_VISIT(self->bound_type);
+    Py_VISIT(self->parent);
 
     return 0;
 }
@@ -1500,6 +1514,7 @@ static int WraptFunctionWrapperBase_clear(WraptFunctionWrapperObject *self)
     Py_CLEAR(self->wrapper);
     Py_CLEAR(self->adapter);
     Py_CLEAR(self->bound_type);
+    Py_CLEAR(self->parent);
 
     return 0;
 }
@@ -1561,8 +1576,8 @@ static PyObject *WraptFunctionWrapperBase_descr_get(
         type = Py_None;
 
     if (descriptor) {
-        result = PyObject_CallFunction(self->bound_type, "(OOO)",
-                descriptor, obj, self->wrapper);
+        result = PyObject_CallFunction(self->bound_type, "(OOOOOO)",
+                descriptor, obj, self->wrapper, self->adapter, Py_None, self);
     }
 
     Py_XDECREF(descriptor);
@@ -1690,6 +1705,20 @@ static PyObject *WraptFunctionWrapperBase_get_self_bound_type(
     return self->bound_type;
 }
 
+/* ------------------------------------------------------------------------- */
+
+static PyObject *WraptFunctionWrapperBase_get_self_parent(
+        WraptFunctionWrapperObject *self, void *closure)
+{
+    if (!self->parent) {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+
+    Py_INCREF(self->parent);
+    return self->parent;
+}
+
 /* ------------------------------------------------------------------------- */;
 
 static PyGetSetDef WraptFunctionWrapperBase_getset[] = {
@@ -1719,6 +1748,8 @@ static PyGetSetDef WraptFunctionWrapperBase_getset[] = {
                             NULL, 0 },
     { "_self_bound_type",   (getter)WraptFunctionWrapperBase_get_self_bound_type,
                             NULL, 0 },
+    { "_self_parent",       (getter)WraptFunctionWrapperBase_get_self_parent,
+                            NULL, 0 },
     { NULL },
 };
 
@@ -2037,7 +2068,7 @@ static int WraptFunctionWrapper_init(WraptFunctionWrapperObject *self,
         bound_type = (PyObject *)&WraptBoundMethodWrapper_Type;
 
     result = WraptFunctionWrapperBase_raw_init(self, wrapped, Py_None,
-            wrapper, adapter, bound_type);
+            wrapper, adapter, bound_type, Py_None);
 
     return result;
 }
